// Draw maze
//
// Usage: draw [-dsgx] < data
//
//   -d  debug
//   -s  solve the maze (graphics only)
//   -g  graphics output
//   -x  "XX" style output
//
// Default output is "L" style using | _ characters
//
// R. Perry, March 1991; updated Nov 2006, Jan 2016, Apr 2023
//
#include <stdio.h>
#include "maze.h"

// graphics output
//
void maze_print_G( struct cell **maze, int m, int n)
{
  printf("\n%d %d\n%d %d\n", 0+1, m+1, n+1, m+1); // top edge

  for( int i = 0; i < m; i++)
  {	
    for( int j = 0; j < n; j++)
    {
      if( (i != 0 || j != 0) && maze[i][j].left == CLOSED)
	printf("\n%d %d\n%d %d\n", j+1, m-i+1, j+1, m-i-1+1);

      if( maze[i][j].down == CLOSED)
	printf("\n%d %d\n%d %d\n", j+1, m-i-1+1, j+1+1, m-i-1+1);
    }

    if( i != m-1) /* not maze exit */
      printf("\n%d %d\n%d %d\n", n+1, m-i+1, n+1, m-i-1+1);
  }
}

// display solution
//
void maze_print_G_solve( struct cell **maze, int m, int n)
{
  for( int i = 0; i < m; i++)
  {	
    for( int j = 0; j < n; j++)
    {
      if( maze[i][j].path & LEFT)
	printf( "\n%g %g\n%g %g\n", j+1-0.5, m-i-1+1+0.5, j+1+1-0.5, m-i-1+1+0.5);

      if( maze[i][j].path & DOWN)
	printf( "\n%g %g\n%g %g\n", j+1+0.5, m-i+1-0.5, j+1+0.5, m-i-1+1-0.5);
    }
  }
}

// "L" style output using | _ characters
//
void maze_print_L( struct cell **maze, int m, int n)
{
  /* top wall */

  printf( " _");
  for( int j = 0; j < n-1; ++j) printf( "__"); 
  putchar( '\n');

  for( int i = 0; i < m; ++i)
  {
    for( int j = 0; j < n; ++j)
    {
      if( maze[i][j].left == CLOSED) putchar( '|'); else putchar( ' ');
      if( maze[i][j].down == CLOSED) putchar( '_'); else putchar( ' ');
    }

    if( i == m-1) /* maze exit */
      printf( " \n");
    else
      printf( "|\n");
  }
}

// compute maze solution
//
int maze_solve( struct cell **maze, int m, int n, int i, int j)
{
  if( i == m-1 && j == n-1) /* success, found the exit */
    return 1;
  else if( maze[i][j].path )
    return 0;

  maze[i][j].path = ON;

  if( j > 0 && maze[i][j].left == OPEN && maze_solve( maze, m, n, i, j-1) ) /* left */
  {
    maze[i][j].path |= LEFT; return 1;
  }
  else if( i < m-1 && maze[i][j].down == OPEN && maze_solve( maze, m, n, i+1, j) ) /* down */
  {
    maze[i][j].path |= DOWN; return 1;
  }
  else if( i > 0 && maze[i-1][j].down == OPEN && maze_solve( maze, m, n, i-1, j) ) /* up */
  {
    maze[i-1][j].path |= DOWN; return 1;
  }
  else if( j < n-1 && maze[i][j+1].left == OPEN && maze_solve( maze, m, n, i, j+1) ) /* right */
  {
    maze[i][j+1].path |= LEFT; return 1;
  }
  else /* maze[i][j] is not on the solution path */
  {
    maze[i][j].path = OFF; return 0;
  }
}

int main( int argc, char *argv[])
{
  int m, n;  Maze_fn maze_print = maze_print_L;

  while( argc > 1 && argv[1][0] == '-')
  {
    switch( argv[1][1])
    {
      case 'D': case 'd': debug = 1; break;
      case 'G': case 'g': maze_print = maze_print_G; break;
      case 'S': case 's': solve = 1; break;
      case 'X': case 'x': maze_print = maze_print_X; break;
      default: fprintf( stderr, "maze: unknown option: %s\n", argv[1]); break;
    }
    --argc, ++argv;
  }

  if( scanf( "%d%d", &m, &n) != 2) { fprintf( stderr, "maze: bad input size\n"); return 1; }

  struct cell **maze = maze_alloc( m, n);

  if( maze == 0) { fprintf( stderr, "maze_alloc() failed.\n"); return 1; }

  maze_init( maze, m, n);

  if( maze_read( maze, m, n) == 0) { fprintf( stderr, "maze_read() failed.\n"); return 1; }

  if( maze_print == maze_print_G && solve)
  {
    maze_solve( maze, m, n, 0, 0);
    maze_print_G_solve( maze, m, n);
  }
  else
  {
    maze_print( maze, m, n);
  }

  return 0;
}
